function [p,t] = halcyon(y_original,fs_original,param)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% halcyon: Parametrized implementation of halcyon()
% This script implements HALCYON: instantaneous pitch estimator based on
% instantaneous phase detection on a multi-rate scheme. Original code can 
% be found in: 
%   Webpage - http://dsp.tut.su/halcyon.html 
%   Contact - E. Azarov (azarov@bsuir.by)
%
% Author: Manuel Diaz Cadiz (mdiazcad@bu.edu)
%         SteppLab, Boston University, MA
% Date:   04/10/2018
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
disp('Begin f_o estimation...');
% disp('> param:');
% disp(param);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% PARAMETERS ------ Default | Description --------------------------------
min_f0 = param.min_f0; %  50 | Min. f0 candidate
max_f0 = param.max_f0; % 450 | Max. f0 candidate
N_cndt = param.N_cndt; % 100 | Num. f0 candidates 
L = param.L;           %   4 | Num. cycles within the time window 
K = param.K;           %   8 | Num. harm. used = Num. of filters in bank 
V = param.V;           %   1 | Num. shifts for instant phase estimates (>0)
fstep = param.fstep;   %   8 | Frame input step (controls time resolution)
dpstep = param.dpstep; %  21 | Dynamic prog. step (num. samples to jump).
R = (2*K+1);           % 2K+1| Candidate2SampleRate multiplier.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Adjust fs_input signal to max. sampling rate used (R*max_f0)  
fs_input = R*max_f0; 
y = resample(y_original,fs_input,fs_original); 
y = y(:);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% F0 candidates and respective Sampling freq.
f_candidate_all = 2.^linspace(log2(min_f0),log2(max_f0),N_cndt);
f_samping_all   = R*f_candidate_all; 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% N_in:  Frame size window needed to capture biggest time segment: 
%        L signal cycles for lowest f0 (min_f0) at fs_input.
N_in  = floor((round(L*fs_input/min_f0))/2)*2+1;  % Also, forced to be odd.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% N_out: Frame size window to capture exact L candidate cycles for each 
%        sampling rate. Since f_sampling/f_candidate = R, N_out is const.
%        N_out consider as well extra samples for prod() operation
%        between sample-shifted responses (2*V+1 lags)
N_out = floor(( L*R + 2*V+1  )/2)*2+1;            % Also, forced to be odd.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% N_FFT: Freq. frame size to calculate fft(). 
%        Same as N_out, but extra samples must be removed to calculate
%        instantaneous phase estimates from 2*V+1 consecutive time frames
N_FFT = floor((     L*R      )/2)*2;             % Also, forced to be even.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% fft_bins: FFT bins at candidate locations. They are function of L and K
fft_bins = (1:K)*L + 1;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Re-sampling matrix generation
if isfield(param,'SINC_MTX')
    SINC_MTX = param.SINC_MTX;
else
    disp('Sampling matrix not found, generating new SINC_MTX...');
    alpha = 1; % Internal param. controls BW of the samplers (default = 1)
    SINC_MTX = gen_shannon_mtx(fs_input,f_samping_all,N_in,N_out,alpha); 
    disp('Sampling matrix generated.');
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Window def. for re-sampled frames
wind=fir1(N_FFT-1,1/N_FFT*2,'noscale')';
winn_GRID_MTX  = repmat(wind,1,N_cndt);
% Harmonic idx grid for W_MTX division 
[k_GRID_MTX,f_sample_MTX] = ndgrid(1:K,f_samping_all);
% Equalization ramp fro low-freq damping 
Eqz_GRID_MTX   = repmat(linspace(0.6,1,N_cndt),2*V+1,1);
% Transform factors from W_MTX (harmonics on [rad]) to F_MTX (F0s on [Hz])
W2f_GRID_MTX   = f_sample_MTX./k_GRID_MTX/(2*pi);
% Frame time progression vector at fstep
frame_idx_all  = 1:fstep:numel(y)-N_in+1;
N_frame_idx    = numel(frame_idx_all);
% Pitch candidate generator function (PCGF) and F0_surf allocation
PCGF   = zeros(N_cndt,N_frame_idx);
F0surf = zeros(N_cndt,N_frame_idx);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% HALCYON MAIN LOOP
disp('Calculating...');
for idx=1:N_frame_idx
    frame  = y(frame_idx_all(idx):frame_idx_all(idx)+N_in-1);  
    framev = SINC_MTX*frame;   
    y_MTX  = reshape(framev,N_out,N_cndt);      
    Y_MTX_STK = zeros(2*V+2,K,N_cndt);
    for j=1:2*V+2
        this_y_MTX = y_MTX(j:N_FFT+j-1,:);
        this_y_MTX = this_y_MTX.*winn_GRID_MTX;
        this_y_MTX = this_y_MTX./...
                     repmat(max(sqrt(sum(this_y_MTX.^2,1)),eps),N_FFT,1);
        Y_MTX = fft(this_y_MTX,[],1);
        Y_MTX = Y_MTX(fft_bins,:);
        Y_MTX_STK(j,:,:) = Y_MTX;
    end 
    P_MTX_STK = angle(Y_MTX_STK);
    PCGF_STK  = zeros(2*V+1,N_cndt);
    for j=1:2*V+1      
        P_MTX_STK(j+1,:,:) = P_MTX_STK(j+1,:,:) - ...
                 2*pi*round((P_MTX_STK(j+1,:,:)-P_MTX_STK(j,:,:))/2/pi);   
        W_MTX = squeeze( P_MTX_STK(j+1,:,:)- P_MTX_STK(j,:,:) );
        A_MTX = squeeze( abs(Y_MTX_STK(j,:,:)) );     
        PCGF_STK(j,:) = sum( A_MTX.*cos(R*W_MTX) ,1);
        if j == V+2
             W_MTX_mid = W_MTX;
             A_MTX_mid = A_MTX; 
        end
    end
    this_PCGF     = prod(PCGF_STK.*Eqz_GRID_MTX,1);
    PCGF(:,idx)   = this_PCGF'/max(this_PCGF); 
    F0surf(:,idx) = ( sum(W2f_GRID_MTX.*W_MTX_mid.*(A_MTX_mid.^2),1)./...
                      sum(A_MTX_mid.^2,1) )';
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Dynamic programing (min path cost.)
[~,fc_path]  = dp(-PCGF',dpstep);
pc_path1 = zeros(1,N_frame_idx);
for i=1:N_frame_idx
    pc_path1(i) = F0surf(fc_path(i),i);
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Alternative pitch estimates code ('others' variables, do not use yet...)
% Define 5% bands around coarse pitch candidates
% fc_band1 = max(  1   , fc_path - round(N_cndt/20) );
% fc_band2 = min(N_cndt, fc_path + round(N_cndt/20) );
% % Calculate coarse pitch estimates with those paths
% pc_coarse   = f_candidate_all(fc_path);
% pc_coarseb1 = f_candidate_all(fc_band1);
% pc_coarseb2 = f_candidate_all(fc_band2);
% % Refine solution by PCGF contour around pc_path estimate.
% pc_path2  = zeros(1,N_frame_idx);
% for i=1:N_frame_idx
%     band_idx  = fc_band1(i):fc_band2(i);
%     PCGF_crop = PCGF(band_idx,i);
%     pc_path2(i) = f_candidate_all(band_idx)*PCGF_crop/sum(PCGF_crop);
% end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Main outputs
p  = pc_path1;
t  = (frame_idx_all-1+(N_in-1)/2)/fs_input;
% pc = f_candidate_all;
% Others
% others.pc_path1    = pc_path1;
% others.pc_path2    = pc_path2;
% others.pc_coarse   = pc_coarse;
% others.pc_coarseb1 = pc_coarseb1;
% others.pc_coarseb2 = pc_coarseb2;
% others.fc_path     = fc_path;
% others.fc_band1    = fc_band1;
% others.fc_band2    = fc_band2;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
disp('f_o estimation complete.')
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Resampling matrix (Shannon-Toeplitz) function generator
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [SINC_MTX] = gen_shannon_mtx(fs_in,fs_out_vec,N_in,N_out,alpha)
    N_cndt = numel(fs_out_vec);    
    HF_N_offset_in  = (N_in-1)/2;
    HF_N_offset_out = (N_out-1)/2;
    hamm_win_GRID = repmat(hamming(N_in)',N_out,1);
    SINC_MTX      = zeros(N_out*N_cndt,N_in);
    offset = 0; 
    for j=1:N_cndt
        fs_out = fs_out_vec(j);
        w_bw  = alpha*(2*pi)*(fs_out/fs_in);
        t_in  = (-HF_N_offset_in +(0:N_in-1) )/fs_in;  n_in  = fs_in*t_in;
        t_out = (-HF_N_offset_out+(0:N_out-1))/fs_out; n_out = fs_in*t_out;
        [n_out_GRID,n_in_GRID] = ndgrid(n_out,n_in);
        temp = (w_bw/2/pi).*sinc( w_bw/2/pi.*(n_out_GRID - n_in_GRID) );
        temp = temp.*hamm_win_GRID;
        temp = temp./repmat(sum(temp,2),1,N_in);
        SINC_MTX(1+offset:N_out+offset,:) = temp;
        offset = N_out + offset;
    end
end

function [p,q,D_vec,phi]=dp(Value_matrix,Max_step)

[r,c] = size(Value_matrix);

D = zeros(r+1, c+Max_step*2);
D(1,:) = 0;
D(:,[1:Max_step Max_step+c+1:c+Max_step*2]) = NaN;
D(2:(r+1), Max_step+1:Max_step+c) = Value_matrix;

phi = zeros(r,c); %Return indexes

for i = 1:r; 
  for j = Max_step+1:Max_step+c;
    [dmax, tb] = min(D(i, j-Max_step:j+Max_step));
    D(i+1,j) = D(i+1,j)+dmax;
    phi(i,j-Max_step) = tb;
  end
end

%Retrieving the best path
i = r; 
[~,j] = min(D(r+1,Max_step+1:Max_step+c));
p = i;
q = j;
while i > 1 && j >= 1
  tb = phi(i,j);

  i=i-1;
  j=j-(Max_step+1)+tb;
  
  if(j==0)
      break;
  end
  
  p = [i,p];
  q = [j,q];
end

D_vec = D(r+1,:);
end

